from .fem.mesh import FemMesh, SolidMesh
from .fem.pointdata import PointData
from .fem.structure import Structure
from .fem.linemesh import LineMesh
from .fem.surfacemesh import SurfaceMesh
from .material.beam.bernoulli.section import BeamSection

__version__ = "0.0.32"

__description__ = "High-Performance Computational Mechanics in Python."
